

#include "main.h"

#include "glwrap.h"
#include "glut.h"
#include "myext.h"

#include <string.h>
#include <stdio.h>



//fixme:
//  could put Skybox in its own header/source file.
// ==hplus==

class Skybox {
public:
  Skybox();
  void release();
  bool load( char const * name );
  void draw();
private:
  ~Skybox();
  GWState sides_[ 6 ];
  static float coords[ 6 ][ 4 ][ 3 ];
  static float uvs[ 6 ][ 4 ][ 2 ];
  static char const * suffices[ 6 ];
};

//  Typically, you'd have artists create geometry as models. 
//  However, a skybox is simple enough that I don't want to code 
//  up a model loader just for this example. Thus, here's the 
//  geometry, all inline. I'm putting it 9 units from the origin 
//  to make sure it doesn't clip through our near clipping plane.
//  Other than that, it's really quite arbitrary where we put it 
//  as long as it's centered on the origin.
float Skybox::coords[ 6 ][ 4 ][ 3 ] =
{
  {   //  Z
    { -9, -9, 9, },
    { -9, 9, 9, },
    { 9, 9, 9, },
    { 9, -9, 9, },
  },
  {   //  X
    { 9, -9, 9, },
    { 9, 9, 9, },
    { 9, 9, -9, },
    { 9, -9, -9, },
  },
  {   //  -Z
    { 9, -9, -9, },
    { 9, 9, -9, },
    { -9, 9, -9, },
    { -9, -9, -9, },
  },
  {   //  -X
    { -9, -9, -9, },
    { -9, 9, -9, },
    { -9, 9, 9, },
    { -9, -9, 9, },
  },
  {   //  Y
    { -9, 9, 9, },
    { -9, 9, -9, },
    { 9, 9, -9, },
    { 9, 9, 9, },
  },
  {   //  -Y
    { -9, -9, -9, },
    { -9, -9, 9, },
    { 9, -9, 9, },
    { 9, -9, -9, },
  },
};

//  It's important to get the u/v coordinates right for each 
//  vertex for each texture, so the cube matches up once it's 
//  drawn.
float Skybox::uvs[ 6 ][ 4 ][ 2 ] =
{
  {   //  Z
    { 1, 0, },
    { 1, 1, },
    { 0, 1, },
    { 0, 0, },
  },
  {   //  X
    { 1, 0, },
    { 1, 1, },
    { 0, 1, },
    { 0, 0, },
  },
  {   //  -Z
    { 1, 0, },
    { 1, 1, },
    { 0, 1, },
    { 0, 0, },
  },
  {   //  -X
    { 1, 0, },
    { 1, 1, },
    { 0, 1, },
    { 0, 0, },
  },
  {   //  Y
    { 0, 0, },
    { 1, 0, },
    { 1, 1, },
    { 0, 1, },
  },
  {   //  -Y
    { 1, 1, },
    { 0, 1, },
    { 0, 0, },
    { 1, 0, },
  },
};

char const * Skybox::suffices[ 6 ] = {
  "_pz",
  "_px",
  "_nz",
  "_nx",
  "_py",
  "_ny",
};


Skybox::Skybox()
{
}

void Skybox::release()
{
  delete this;
}

bool Skybox::load( char const * name )
{
  char path[ PathLengthMax ];
  bool ok = true;
  for( unsigned int ix = 0; ix < 6; ++ix ) {
    sprintf( path, "%s%s.tga", name, suffices[ ix ] );
    sides_[ ix ].texture = TextureRef( path, true );
    if( !sides_[ ix ].texture.consistent() ) {
      error( "cannot load texture: %s", path );
      ok = false;
    }
  }
  return ok;
}

void Skybox::draw()
{
  //  The skybox needs to be drawn centered on the origin of the 
  //  eye, no matter what the "world" transform you're using may 
  //  be. This assert makes sure that such is the case. It also 
  //  makes sure there's no skew. It could be tightened up to 
  //  also make sure there's no non-uniform scale...
#if !defined( NDEBUG )
  float mx[ 16 ];
  glGetFloatv( GL_MODELVIEW_MATRIX, mx );
  assert( mx[12] == 0 && mx[13] == 0 && mx[14] == 0 );
  assert( mx[3] == 0 && mx[7] == 0 && mx[11] == 0 );
  assert( mx[15] == 1 );
#endif
  for( unsigned int ix = 0; ix < 6; ++ix ) {
    sides_[ ix ].draw( GL_QUADS, NULL, 4, (float *)coords[ ix ], (float *)uvs[ ix ], NULL, NULL );
  }
}

Skybox::~Skybox()
{
}


Skybox * gSkybox;

void initSkybox()
{
  gSkybox = new Skybox;
  gSkybox->load( "skybox" );
}


float gYaw;
float gPitch;



struct IntOptionValue {
  char const * name;
  int val;
};
static IntOptionValue intValues[] = {
  { "winSizeX", WINWIDTH },
  { "winSizeY", WINHEIGHT },
};

int
intOption( char const * o )
{
  for( int ix = 0; ix < sizeof(intValues)/sizeof(intValues[0]); ++ix ) {
    if( !_stricmp( intValues[ix].name, o ) ) {
      return intValues[ix].val;
    }
  }
  assert( 0 );
  return -1;
}

static void
vanillaGl( )
{
  //  make sure there's no leakage between frames
  gwReset( );

  //  set up our basic projection
  glMatrixMode( GL_PROJECTION );
  glLoadIdentity();
  float fw = float(intValues[1].val)/float(intValues[0].val);
  glFrustum( -1, 1, -fw, fw, 1, 1000 );
  glMatrixMode( GL_MODELVIEW );

  assert( !glGetError( ) );
}




void display()
{
  vanillaGl();

  glClear( GL_DEPTH_BUFFER_BIT ); //  we don't need to clear color, because we will draw
                                  //  over the entire screen

  glRotatef( gPitch * 180 / M_PI, 1, 0, 0 );
  glRotatef( gYaw * 180 / M_PI, 0, -1, 0 );
  
  gSkybox->draw();

  gwSwapBuffers( );
}


bool options[ 128 ];

static void
putOptions( int argc, char ** argv )
{
  for( int ix = 0; ix < argc; ++ix ) {
    if( argv[ix][0] == '-' && argv[ix][1] != '-' ) {
      char * p = &argv[ix][1];
      while( *p ) {
        options[*p & 0x7f] = true;
        ++p;
      }
    }
    else {
      char * eq = strchr( argv[ ix ], '=' );
      if( eq ) {
        *eq = 0;
        ++eq;
        for( int o = 0; o < sizeof(intValues)/sizeof(intValues[0]); ++o ) {
          if( !stricmp( intValues[o].name, argv[ix] ) ) {
            intValues[o].val = atoi(eq);
            debug( "setting %s to %d\n", intValues[o].name, intValues[o].val );
            goto foundit;
          }
        }
      }
      warning( "unknown option: %s\n", argv[ix] );
foundit:
      ;
    }
  }
}

bool
hasOption( char ch )
{
  return options[ ch & 0x7f ];
}

void keyboard( unsigned char key, int x, int y )
{
  if( key == 'q' || key == 'Q' || key == 27 ) {
    exit(0);
  }
}

void special( int key, int x, int y )
{
  if( key == GLUT_KEY_RIGHT ) {
    gYaw -= 0.1f;
  }
  else if( key == GLUT_KEY_LEFT ) {
    gYaw += 0.1f;
  }
  else if( key == GLUT_KEY_UP ) {
    gPitch -= 0.05f; 
  }
  else if( key == GLUT_KEY_DOWN ) {
    gPitch += 0.05f; 
  }

  if( gPitch > M_PI ) {
    gPitch -= 2*M_PI;
  }
  else if( gPitch < -M_PI ) {
    gPitch += 2*M_PI;
  }
  if( gYaw > M_PI ) {
    gYaw -= 2*M_PI;
  }
  else if( gYaw < -M_PI ) {
    gYaw += 2*M_PI;
  }
}

void timer( )
{
  glutPostRedisplay();  //  this is maybe not the most efficient...
}


void mouse( int button, int state, int x, int y )
{
  // state == GLUT_DOWN ? true : false
}

void activeMotion( int x, int y )
{
}

void passiveMotion( int x, int y )
{
}

void
goBackToNormalScreen( )
{
#if defined( WIN32 )
  ChangeDisplaySettings( 0, 0 );
#endif
}

int main(int argc, char** argv)
{
#if !defined( NDEBUG )
  initLog( "log.txt", LogFilterNone );
#else
  initLog( "log.txt", LogFilterDebug );
#endif

#if defined( WIN32 )
  RECT r = { 0, 0, 0, 0 };
  AdjustWindowRectEx( &r, WS_OVERLAPPEDWINDOW, false, WS_EX_OVERLAPPEDWINDOW );
#endif
  
  putOptions( argc, argv );
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
  glutInitWindowSize(intOption("winSizeX"),intOption("winSizeY"));
  if( !options['w'] ) {
// this doesn't seem to work -- can't find documentation anywhere
//    glutGameModeString( "800x600@60" );
//    glutEnterGameMode( );
    glutInitWindowPosition(0, 0);
  }
  else {
    glutInitWindowPosition(100, 0);
  }
  glutCreateWindow("Main Window");
  glutSetKeyRepeat( GLUT_KEY_REPEAT_OFF );
  if( !options[ 'w' ] ) {
#if defined( WIN32 )
    // not windowed -- go fullscreen
    DEVMODE dm;
    memset( &dm, 0, sizeof( dm ) );
    dm.dmSize = sizeof( DEVMODE );
    dm.dmFields = DM_PELSWIDTH | DM_PELSHEIGHT;
    dm.dmBitsPerPel = 32;
    dm.dmPelsWidth = intOption("winSizeX");
    dm.dmPelsHeight = intOption("winSizeY");
    dm.dmDisplayFlags = 0;
    dm.dmDisplayFrequency = 60;
    BOOL b = (ChangeDisplaySettings( &dm, CDS_FULLSCREEN ) == DISP_CHANGE_SUCCESSFUL);
    if( !b ) {
      error( "ChangeDisplaySettings(): cannot go fullscreen!\n" );
    }
#endif
    atexit( goBackToNormalScreen );
    glutPositionWindow( 0, 0 );
    glutReshapeWindow( intOption("winSizeX"), intOption("winSizeY") );
  }
  vanillaGl( );
  gwFlushState( );
  glutDisplayFunc( display );
  glutKeyboardFunc( keyboard );
  glutSpecialFunc( special );
  glutIdleFunc( timer );
  glutMouseFunc( mouse );
  glutMotionFunc( activeMotion );
  glutPassiveMotionFunc( passiveMotion );

  initSkybox();

  glutMainLoop();

  return 0;
}


#if defined( WIN32 )
static int
splitCmdLine( char const * cmdLine, char ** argv, int n )
{
  int args = 0;
  while( n > 1 ) {
    while( *cmdLine && isspace( *cmdLine ) ) {
      ++cmdLine;
    }
    if( !*cmdLine ) break;
    char const * lStart;
    char const * lEnd;
    if( *cmdLine == '\"' ) {
      lStart = cmdLine+1;
      lEnd = strchr( lStart, '\"' );
    }
    else {
      lStart = cmdLine;
      lEnd = strchr( lStart, ' ' );
    }
    if( !lEnd ) {
      lEnd = lStart + strlen( lStart );
    }
    *argv = new char[ lEnd-lStart+1 ];
    memcpy( *argv, lStart, lEnd-lStart );
    (*argv)[ lEnd-lStart ] = 0;
    --n;
    ++args;
    ++argv;
    if( *lEnd ) {
      cmdLine = lEnd+1;
    }
    else {
      cmdLine = lEnd;
    }
  }
  *argv = 0;
  return args;
}

int __stdcall
WinMain(
  HINSTANCE cur,
  HINSTANCE prev,
  LPSTR     cmdLine,
  int       cmdShow )
{
  char * argv[ 100 ];
  char progName[ 32 ] = "skybox";
  argv[ 0 ] = progName;
  int argc = 1 + splitCmdLine( cmdLine, &argv[ 1 ], 98 );
  return main( argc, argv );
}

int 
strcasecmp( char const * a, char const * b )
{
  while( *a && *b && (tolower(*a) == tolower(*b)) ) {
    ++a;
    ++b;
  }
  return tolower(*a) - tolower(*b);
}
#endif



//  FIXME: move log into its own file
//  ==hplus==


static unsigned int s_logFilterLevel;
static FILE * s_logFile;

void
initLog( char const * fileName, unsigned int level )
{
  if( s_logFile ) {
    fclose( s_logFile );
  }
  s_logFile = fopen( fileName, "wb" );
  assert( s_logFile ); // crappy error reporting
}

static void
logOutput( unsigned int level, char const * fmt, va_list l )
{
  if( !s_logFile ) {
    return;
  }
  if( level <= s_logFilterLevel ) {
    return;
  }
  vfprintf( s_logFile, fmt, l );
  fflush( s_logFile );
}

void debug( char const * fmt, ... )
{
  va_list vl;
  va_start( vl, fmt );
  logOutput( LogFilterDebug, fmt, vl );
  va_end( vl );
}

void warning( char const * fmt, ... )
{
  va_list vl;
  va_start( vl, fmt );
  logOutput( LogFilterWarning, fmt, vl );
  va_end( vl );
}

void error( char const * fmt, ... )
{
  assert( 0 ); // catch errors as they happen!
  va_list vl;
  va_start( vl, fmt );
  logOutput( LogFilterError, fmt, vl );
  va_end( vl );
}

